var bg = (function (exports) {
    'use strict';

    //
    //  Copyright Yahoo 2021
    //
    //  Derived from tinyqueue, Copyright 2017 Vladamir Agafonkin, ISC License, https://github.com/mourner/tinyqueue
    //
    class BufferedPriorityQueue {
        /*
        A binary heap implementation that works for specifically for behaviors.
        Delays heapifying new elements until the next pop. Adding multiple extents
        in the same behavior can lead to reheaping. The delay prevents behaviors
        from being resorted multiple times.
         */
        queue = [];
        buffer = [];
        constructor() {
        }
        push(behavior) {
            this.buffer.push(behavior);
        }
        peek() {
            if (this.length == 0) {
                return undefined;
            }
            else {
                this.heapifyBuffer();
                return this.queue[0];
            }
        }
        pop() {
            if (this.length == 0) {
                return undefined;
            }
            else {
                this.heapifyBuffer();
                // The min is the first one
                // We want to pop from the front (shift) but that is
                // expensive in js so we take the last element and put
                // it at the top and sort it back down (pop is cheap)
                let b = this.queue[0];
                let last = this.queue.pop();
                if (this.queue.length > 0) {
                    this.queue[0] = last;
                    this.down(0);
                }
                return b;
            }
        }
        heapifyBuffer() {
            // heapify elements in the buffer
            for (let i = 0; i < this.buffer.length; i++) {
                let b = this.buffer[i];
                this.queue.push(b);
                if (this.queue.length > 1) {
                    this.up(this.queue.length - 1);
                }
            }
            this.buffer.length = 0;
        }
        get length() {
            return this.queue.length + this.buffer.length;
        }
        unsort() {
            this.buffer = this.buffer.concat(this.queue);
            this.queue.length = 0;
        }
        clear() {
            this.queue.length = 0;
            this.buffer.length = 0;
        }
        up(pos) {
            let b_up = this.queue[pos];
            while (pos > 0) {
                let parent = (pos - 1) >> 1;
                let b_parent = this.queue[parent];
                if (b_up.order >= b_parent.order) {
                    break;
                }
                this.queue[pos] = b_parent;
                pos = parent;
            }
            this.queue[pos] = b_up;
        }
        down(pos) {
            const halfLength = this.queue.length >> 1;
            const b_down = this.queue[pos];
            while (pos < halfLength) {
                let left = (pos << 1) + 1;
                let best = this.queue[left];
                const right = left + 1;
                if (right < this.queue.length && this.queue[right].order < best.order) {
                    left = right;
                    best = this.queue[right];
                }
                if (best.order >= b_down.order) {
                    break;
                }
                this.queue[pos] = best;
                pos = left;
            }
            this.queue[pos] = b_down;
        }
    }

    var OrderingState;
    (function (OrderingState) {
        OrderingState[OrderingState["Untracked"] = 0] = "Untracked";
        OrderingState[OrderingState["NeedsOrdering"] = 1] = "NeedsOrdering";
        OrderingState[OrderingState["Clearing"] = 2] = "Clearing";
        OrderingState[OrderingState["Ordering"] = 3] = "Ordering";
        OrderingState[OrderingState["Ordered"] = 4] = "Ordered"; // has a valid order
    })(OrderingState || (OrderingState = {}));
    exports.RelinkingOrder = void 0;
    (function (RelinkingOrder) {
        RelinkingOrder[RelinkingOrder["relinkingOrderPrior"] = 0] = "relinkingOrderPrior";
        RelinkingOrder[RelinkingOrder["relinkingOrderSubsequent"] = 1] = "relinkingOrderSubsequent";
    })(exports.RelinkingOrder || (exports.RelinkingOrder = {}));
    exports.RelinkingTarget = void 0;
    (function (RelinkingTarget) {
        RelinkingTarget[RelinkingTarget["demand"] = 0] = "demand";
        RelinkingTarget[RelinkingTarget["supply"] = 1] = "supply";
    })(exports.RelinkingTarget || (exports.RelinkingTarget = {}));
    exports.ResourceType = void 0;
    (function (ResourceType) {
        ResourceType[ResourceType["resource"] = 0] = "resource";
        ResourceType[ResourceType["moment"] = 1] = "moment";
        ResourceType[ResourceType["typedMoment"] = 2] = "typedMoment";
        ResourceType[ResourceType["state"] = 3] = "state";
    })(exports.ResourceType || (exports.ResourceType = {}));
    exports.LinkType = void 0;
    (function (LinkType) {
        LinkType[LinkType["reactive"] = 0] = "reactive";
        LinkType[LinkType["order"] = 1] = "order";
    })(exports.LinkType || (exports.LinkType = {}));
    class GraphEvent {
        sequence;
        timestamp;
        static initialEvent = new GraphEvent(0, new Date(0));
        constructor(sequence, timestamp) {
            this.sequence = sequence;
            this.timestamp = timestamp;
        }
    }

    //
    var LinkType;
    (function (LinkType) {
        LinkType[LinkType["reactive"] = 0] = "reactive";
        LinkType[LinkType["order"] = 1] = "order";
    })(LinkType || (LinkType = {}));
    class Resource {
        debugName;
        isResource = true;
        extent;
        graph;
        subsequents = new Set();
        suppliedBy = null;
        skipChecks = false;
        didUpdateSubscribers;
        constructor(extent, name) {
            this.extent = extent;
            this.graph = extent.graph;
            extent.addResource(this);
            if (name !== undefined) {
                this.debugName = name;
            }
            else {
                this.debugName = null;
            }
        }
        get order() {
            return { resource: this, type: LinkType.order };
        }
        get resource() {
            return this;
        }
        get type() {
            return LinkType.reactive;
        }
        toString() {
            let name = "Resource";
            if (this.debugName != null) {
                name = this.debugName + "(r)";
            }
            return name;
        }
        assertValidUpdater() {
            let graph = this.graph;
            let currentBehavior = graph.currentBehavior;
            let currentEvent = graph.currentEvent;
            if (currentBehavior == null && currentEvent == null) {
                let err = new Error("Resource must be updated inside a behavior or action.");
                err.resource = this;
                throw err;
            }
            if (this.skipChecks) {
                return;
            }
            if (this.suppliedBy && currentBehavior != this.suppliedBy) {
                let err = new Error("Supplied resource can only be updated by its supplying behavior.");
                err.resource = this;
                err.currentBehavior = currentBehavior;
                throw err;
            }
            if (this.suppliedBy == null && currentBehavior != null) {
                let err = new Error("Unsupplied resource can only be updated in an action.");
                err.resource = this;
                err.currentBehavior = currentBehavior;
                throw err;
            }
        }
        assertValidAccessor() {
            let graph = this.graph;
            let currentBehavior = graph.currentBehavior;
            if (currentBehavior != null && currentBehavior != this.suppliedBy && !currentBehavior.demands?.has(this)) {
                let err = new Error("Cannot access the value or event of a resource inside a behavior unless it is supplied or demanded.");
                err.resource = this;
                err.currentBehavior = currentBehavior;
                throw err;
            }
        }
        get justUpdated() {
            this.assertValidAccessor();
            return false;
        }
        subscribeToJustUpdated(callback) {
            return this._subscribeToJustUpdated({ extent: null, callback: callback });
        }
        _subscribeToJustUpdated(subscription) {
            // returns an unsubscribe callback that caller can call when no longer needed
            if (this.didUpdateSubscribers === undefined) {
                this.didUpdateSubscribers = new Set();
            }
            this.didUpdateSubscribers.add(subscription);
            return (() => {
                this.didUpdateSubscribers.delete(subscription);
            });
        }
        notifyJustUpdatedSubscribers() {
            if (this.didUpdateSubscribers !== undefined && this.didUpdateSubscribers.size > 0) {
                this.graph._notifyJustUpdatedSubscribers(this.didUpdateSubscribers);
            }
        }
    }
    class Moment extends Resource {
        _happened = false;
        _happenedValue = undefined;
        _happenedWhen = null;
        get justUpdated() {
            this.assertValidAccessor();
            return this._happened;
        }
        get value() {
            this.assertValidAccessor();
            return this._happenedValue;
        }
        get event() {
            this.assertValidAccessor();
            return this._happenedWhen;
        }
        toString() {
            let name = "Moment";
            if (this.debugName != null) {
                name = (this.debugName + "(m)");
            }
            if (this._happenedValue !== undefined) {
                name = name + "=" + this._happenedValue;
            }
            if (this._happenedWhen !== null) {
                name = name + " : " + this._happenedWhen.sequence;
            }
            return name;
        }
        justUpdatedTo(value) {
            return this.justUpdated && this._happenedValue == value;
        }
        updateWithAction(value = undefined, debugName) {
            this.graph.action(() => {
                this.update(value);
            }, debugName);
            return;
        }
        update(value = undefined) {
            this.assertValidUpdater();
            this._happened = true;
            this._happenedValue = value;
            this._happenedWhen = this.graph.currentEvent;
            this.notifyJustUpdatedSubscribers();
            this.graph.resourceTouched(this);
            this.graph.trackTransient(this);
        }
        clear() {
            this._happened = false;
            this._happenedValue = undefined;
        }
    }
    class State extends Resource {
        currentState;
        previousState = null;
        constructor(extent, initialState, name) {
            super(extent, name);
            this.currentState = { value: initialState, event: GraphEvent.initialEvent };
        }
        toString() {
            let name = "State";
            if (this.debugName != null) {
                name = (this.debugName + "(s)");
            }
            name = name + "=" + this.currentState.value;
            name = name + " : " + this.currentState.event.sequence;
            return name;
        }
        updateWithAction(newValue, debugName) {
            this.graph.action(() => {
                this.update(newValue);
            }, debugName);
            return;
        }
        update(newValue) {
            if (this.currentState.value === newValue) {
                return;
            }
            this.updateForce(newValue);
        }
        updateForce(newValue) {
            this.assertValidUpdater();
            this._updateForce(newValue);
        }
        _updateForce(newValue) {
            if (this.graph.currentEvent != null && this.currentState.event.sequence < this.graph.currentEvent?.sequence) {
                // captures trace as the value before any updates
                this.previousState = this.currentState;
            }
            this.currentState = { value: newValue, event: this.graph.currentEvent };
            this.notifyJustUpdatedSubscribers();
            this.graph.resourceTouched(this);
            this.graph.trackTransient(this);
        }
        clear() {
            this.previousState = null;
        }
        get value() {
            this.assertValidAccessor();
            return this.currentState.value;
        }
        get event() {
            this.assertValidAccessor();
            return this.currentState.event;
        }
        get trace() {
            if (this.currentState.event === this.graph.currentEvent) {
                return this.previousState;
            }
            else {
                return this.currentState;
            }
        }
        get traceValue() {
            return this.trace.value;
        }
        get traceEvent() {
            return this.trace.event;
        }
        get justUpdated() {
            this.assertValidAccessor();
            return this.currentState.event === this.graph.currentEvent;
        }
        justUpdatedTo(toState) {
            return this.justUpdated && this.currentState.value == toState;
        }
        justUpdatedFrom(fromState) {
            return this.justUpdated && this.traceValue == fromState;
        }
        justUpdatedToFrom(toState, fromState) {
            return this.justUpdatedTo(toState) && this.justUpdatedFrom(fromState);
        }
    }

    //
    const DefaultDateProvider = {
        now: () => {
            return new Date();
        }
    };
    class Graph {
        dateProvider = DefaultDateProvider;
        currentEvent = null;
        lastEvent;
        activatedBehaviors = new BufferedPriorityQueue();
        currentBehavior = null;
        effects = [];
        actions = [];
        untrackedBehaviors = [];
        modifiedDemandBehaviors = [];
        modifiedSupplyBehaviors = [];
        updatedTransients = [];
        needsOrdering = [];
        eventLoopState = null;
        extentsAdded = [];
        extentsRemoved = [];
        validateLifetimes = true;
        justUpdatedCallbacks = new Set();
        constructor() {
            this.lastEvent = GraphEvent.initialEvent;
        }
        action(block, debugName) {
            this.actionHelper({ debugName: debugName, block: block, extent: null, resolve: null });
        }
        actionHelper(action) {
            if (this.eventLoopState != null && (this.eventLoopState.phase == EventLoopPhase.action || this.eventLoopState.phase == EventLoopPhase.updates)) {
                let err = new Error("Action cannot be created directly inside another action or behavior. Consider wrapping it in a side effect block.");
                throw err;
            }
            this.actions.push(action);
            this.eventLoop();
        }
        async actionAsync(block, debugName) {
            return this.actionAsyncHelper({ debugName: debugName, block: block, extent: null, resolve: null });
        }
        async actionAsyncHelper(action) {
            return new Promise((resolve, reject) => {
                try {
                    if (this.eventLoopState != null && (this.eventLoopState.phase == EventLoopPhase.action || this.eventLoopState.phase == EventLoopPhase.updates)) {
                        let err = new Error("Action cannot be created directly inside another action or behavior. Consider wrapping it in a side effect block.");
                        throw err;
                    }
                    action.resolve = resolve;
                    this.actions.push(action);
                    if (this.currentEvent == null) {
                        this.eventLoop();
                    }
                }
                catch (e) {
                    reject(e);
                }
            });
        }
        eventLoop() {
            while (true) {
                try {
                    if (this.activatedBehaviors.length > 0 ||
                        this.untrackedBehaviors.length > 0 ||
                        this.modifiedDemandBehaviors.length > 0 ||
                        this.modifiedSupplyBehaviors.length > 0 ||
                        this.needsOrdering.length > 0) {
                        this.eventLoopState.phase = EventLoopPhase.updates;
                        let sequence = this.currentEvent.sequence;
                        this.addUntrackedBehaviors();
                        this.addUntrackedSupplies();
                        this.addUntrackedDemands(sequence);
                        this.orderBehaviors();
                        this.runNextBehavior(sequence);
                        continue;
                    }
                    if (this.validateLifetimes) {
                        if (this.extentsAdded.length > 0) {
                            this.validateAddedExtents();
                            this.extentsAdded.length = 0;
                        }
                        if (this.extentsRemoved.length > 0) {
                            this.validateRemovedExtents();
                            this.extentsRemoved.length = 0;
                        }
                    }
                    if (this.justUpdatedCallbacks.size > 0) {
                        this.turnSubscriptionsIntoSideEffects();
                    }
                    let effect = this.effects.shift();
                    if (effect) {
                        this.eventLoopState.phase = EventLoopPhase.sideEffects;
                        this.eventLoopState.currentSideEffect = effect;
                        effect.block(effect.extent);
                        if (this.eventLoopState != null) {
                            // side effect could create a synchronous action which would create a nested event loop
                            // which would clear out any existing event loop states
                            this.eventLoopState.currentSideEffect = null;
                        }
                        continue;
                    }
                    if (this.currentEvent) {
                        if (this.eventLoopState.action.resolve != undefined) {
                            this.eventLoopState.action.resolve(undefined);
                        }
                        this.clearTransients();
                        this.lastEvent = this.currentEvent;
                        this.currentEvent = null;
                        this.eventLoopState = null;
                        this.currentBehavior = null;
                    }
                    let action = this.actions.shift();
                    if (action) {
                        let newEvent = new GraphEvent(this.lastEvent.sequence + 1, this.dateProvider.now());
                        this.currentEvent = newEvent;
                        this.eventLoopState = new EventLoopState(action);
                        this.eventLoopState.phase = EventLoopPhase.action;
                        action.block(action.extent);
                        continue;
                    }
                }
                catch (error) {
                    this.currentEvent = null;
                    this.eventLoopState = null;
                    this.actions.length = 0;
                    this.effects.length = 0;
                    this.currentBehavior = null;
                    this.activatedBehaviors.clear();
                    this.justUpdatedCallbacks.clear();
                    this.clearTransients();
                    this.modifiedDemandBehaviors.length = 0;
                    this.modifiedSupplyBehaviors.length = 0;
                    this.untrackedBehaviors.length = 0;
                    this.extentsAdded.length = 0;
                    this.extentsRemoved.length = 0;
                    throw (error);
                }
                // no more tasks so we can exit the event loop
                break;
            }
        }
        validateAddedExtents() {
            // ensure extents with same lifetime also got added
            let needAdding = new Set();
            for (let added of this.extentsAdded) {
                if (added.lifetime != null) {
                    for (let ext of added.lifetime.getAllContainingExtents()) {
                        if (ext.addedToGraphWhen == null) {
                            needAdding.add(ext);
                        }
                    }
                }
            }
            if (needAdding.size > 0) {
                let err = new Error("All extents with unified or parent lifetimes must be added during the same event.");
                err.nonAddedExtents = needAdding;
                throw err;
            }
        }
        validateRemovedExtents() {
            // validate extents with contained lifetimes are also removed
            let needRemoving = new Set();
            for (let removed of this.extentsRemoved) {
                if (removed.lifetime != null) {
                    for (let ext of removed.lifetime.getAllContainedExtents()) {
                        if (ext.addedToGraphWhen != null) {
                            needRemoving.add(ext);
                        }
                    }
                }
            }
            if (needRemoving.size > 0) {
                let err = new Error("All extents with unified or child lifetimes must be removed during the same event.");
                err.nonAddedExtents = needRemoving;
                throw err;
            }
            // validate removed resources are not still linked to remaining behaviors
            for (let removed of this.extentsRemoved) {
                for (let resource of removed.resources) {
                    for (let demandedBy of resource.subsequents) {
                        if (demandedBy.extent.addedToGraphWhen != null) {
                            let err = new Error("Remaining behaviors must remove dynamicDemands to removed resources.");
                            err.remainingBehavior = demandedBy;
                            err.removedResource = resource;
                            throw err;
                        }
                    }
                    if (resource.suppliedBy != null && resource.suppliedBy.extent.addedToGraphWhen != null) {
                        let err = new Error("Remaining behaviors must remove dynamicSupplies to removed resources.");
                        err.remainingBehavior = resource.suppliedBy;
                        err.removedResource = resource;
                        throw err;
                    }
                }
            }
        }
        clearTransients() {
            if (this.updatedTransients.length > 0) {
                for (let transient of this.updatedTransients) {
                    transient.clear();
                }
                this.updatedTransients.length = 0;
            }
        }
        trackTransient(resource) {
            this.updatedTransients.push(resource);
        }
        resourceTouched(resource) {
            if (this.currentEvent != null) {
                if (this.eventLoopState != null && this.eventLoopState.phase == EventLoopPhase.action) {
                    this.eventLoopState.actionUpdates.push(resource);
                }
                for (let subsequent of resource.subsequents) {
                    let isOrderingDemand = subsequent.orderingDemands != null && subsequent.orderingDemands.has(resource);
                    if (!isOrderingDemand) {
                        this.activateBehavior(subsequent, this.currentEvent.sequence);
                    }
                }
            }
        }
        activateBehavior(behavior, sequence) {
            if (behavior.enqueuedWhen == null || behavior.enqueuedWhen < sequence) {
                behavior.enqueuedWhen = sequence;
                this.activatedBehaviors.push(behavior);
            }
        }
        runNextBehavior(sequence) {
            // take top behavior off queue
            let topBehavior = this.activatedBehaviors.pop();
            // if no behavior left we quit
            while (topBehavior !== undefined) {
                if (topBehavior.removedWhen == sequence) {
                    // if this behavior has been removed already then try next one
                    topBehavior = this.activatedBehaviors.pop();
                }
                else {
                    // valid behavior, run it
                    this.currentBehavior = topBehavior;
                    topBehavior.block(topBehavior.extent);
                    this.currentBehavior = null;
                    // check if there is a next behavior and it's the same ordering as the first behavior
                    let nextBehavior = this.activatedBehaviors.peek();
                    if (nextBehavior !== undefined && nextBehavior.order == topBehavior.order) {
                        // if so we will try running it also
                        topBehavior = this.activatedBehaviors.pop();
                    }
                    else {
                        break;
                    }
                }
            }
        }
        subscribeToJustUpdated(resources, callback) {
            return this._subscribeToJustUpdated(resources, { extent: null, callback: callback });
        }
        _subscribeToJustUpdated(resources, subscription) {
            let allUnsubscribes = [];
            for (let resource of resources) {
                let unsubscribe = resource._subscribeToJustUpdated(subscription);
                allUnsubscribes.push(unsubscribe);
            }
            let bigUnsubscribe = () => {
                for (let unsubscribe of allUnsubscribes) {
                    unsubscribe();
                }
            };
            return bigUnsubscribe;
        }
        _notifyJustUpdatedSubscribers(subscribers) {
            subscribers.forEach(subscription => {
                this.justUpdatedCallbacks.add(subscription);
            });
        }
        turnSubscriptionsIntoSideEffects() {
            this.justUpdatedCallbacks.forEach(subscription => {
                // either subscription is not connected to an extent or extent is part of graph
                if (subscription.extent == null || subscription.extent.addedToGraphWhen != null) {
                    this.sideEffectHelper({ block: subscription.callback, behavior: null, extent: subscription.extent });
                }
            });
            this.justUpdatedCallbacks.clear();
        }
        sideEffect(block, debugName) {
            this.sideEffectHelper({ debugName: debugName, block: block, behavior: null, extent: null });
        }
        sideEffectHelper(sideEffect) {
            if (this.currentEvent == null) {
                let err = new Error("Effects can only be added during an event.");
                throw err;
            }
            else if (this.eventLoopState.phase == EventLoopPhase.sideEffects) {
                let err = new Error("Nested side effects don't make sense");
                throw err;
            }
            else {
                this.effects.push(sideEffect);
            }
        }
        debugHere() {
            let text = "";
            if (this.currentEvent != null) {
                text = "Current Event: " + this.currentEvent.sequence + "\n";
                text = text + "Action Updates:\n";
                this.eventLoopState?.actionUpdates.forEach(item => {
                    text = text + " " + item.toString() + "\n";
                });
                if (this.currentBehavior != null) {
                    text = text + "Current Behavior:\n";
                    text = text + this.currentBehavior.toString();
                }
            }
            else {
                text = "No current event.";
            }
            return text;
        }
        addUntrackedBehaviors() {
            if (this.untrackedBehaviors.length > 0) {
                for (let behavior of this.untrackedBehaviors) {
                    this.modifiedDemandBehaviors.push(behavior);
                    this.modifiedSupplyBehaviors.push(behavior);
                }
                this.untrackedBehaviors = [];
            }
        }
        addUntrackedSupplies() {
            if (this.modifiedSupplyBehaviors.length > 0) {
                for (let behavior of this.modifiedSupplyBehaviors) {
                    if (behavior.untrackedSupplies != null) {
                        for (let supply of behavior.untrackedSupplies) {
                            if (this.validateLifetimes && !behavior.extent.hasCompatibleLifetime(supply.extent)) {
                                let err = new Error("Static supplies can only be with extents with the unified or parent lifetimes.");
                                err.currentBehavior = behavior;
                                err.supply = supply;
                                throw err;
                            }
                        }
                    }
                    let allUntrackedSupplies = [...(behavior.untrackedSupplies ?? []), ...(behavior.untrackedDynamicSupplies ?? [])];
                    if (behavior.supplies != null) {
                        for (let existingSupply of behavior.supplies) {
                            existingSupply.suppliedBy = null;
                        }
                    }
                    behavior.supplies = new Set(allUntrackedSupplies);
                    for (let newSupply of behavior.supplies) {
                        if (newSupply.suppliedBy != null && newSupply.suppliedBy != behavior) {
                            let err = new Error("Resource cannot be supplied by more than one behavior.");
                            err.alreadySupplied = newSupply;
                            err.desiredSupplier = behavior;
                            throw err;
                        }
                        newSupply.suppliedBy = behavior;
                    }
                    // technically this behavior doesn't need reordering but its subsequents will
                    // if they already demand a newly supplied resource
                    // setting this to reorder will ensure its subsequents will reorder if needed
                    // in the sortDFS code
                    if (behavior.orderingState != OrderingState.NeedsOrdering) {
                        behavior.orderingState = OrderingState.NeedsOrdering;
                        this.needsOrdering.push(behavior);
                    }
                }
                this.modifiedSupplyBehaviors.length = 0;
            }
        }
        addUntrackedDemands(sequence) {
            if (this.modifiedDemandBehaviors.length > 0) {
                for (let behavior of this.modifiedDemandBehaviors) {
                    if (behavior.untrackedDemands != null) {
                        for (let demand of behavior.untrackedDemands) {
                            if (this.validateLifetimes && !behavior.extent.hasCompatibleLifetime(demand.resource.extent)) {
                                let err = new Error("Static demands can only be with extents with the unified or parent lifetimes.");
                                err.currentBehavior = behavior;
                                err.demand = demand.resource;
                                throw err;
                            }
                        }
                    }
                    let allUntrackedDemands = [...(behavior.untrackedDemands ?? []), ...(behavior.untrackedDynamicDemands ?? [])];
                    let removedDemands;
                    if (behavior.demands != null) {
                        for (let demand of behavior.demands) {
                            if (!allUntrackedDemands.some(linkable => linkable.resource == demand)) {
                                if (removedDemands == undefined) {
                                    removedDemands = [];
                                }
                                removedDemands.push(demand);
                            }
                        }
                    }
                    let addedDemands;
                    for (let linkableDemand of allUntrackedDemands) {
                        let untrackedDemand = linkableDemand.resource;
                        if (untrackedDemand.extent.addedToGraphWhen == null) {
                            let err = new Error("All demands must be added to the graph.");
                            err.currentBehavior = behavior;
                            err.untrackedDemand = untrackedDemand;
                            throw err;
                        }
                        if (behavior.demands == null || !behavior.demands.has(untrackedDemand)) {
                            if (addedDemands == undefined) {
                                addedDemands = [];
                            }
                            addedDemands.push(untrackedDemand);
                        }
                    }
                    let needsRunning = false;
                    if (removedDemands != undefined) {
                        for (let demand of removedDemands) {
                            demand.subsequents.delete(behavior);
                        }
                    }
                    let orderBehavior = behavior.orderingState != OrderingState.Ordered;
                    if (addedDemands != undefined) {
                        for (let demand of addedDemands) {
                            demand.subsequents.add(behavior);
                            if (demand.justUpdated) {
                                needsRunning = true;
                            }
                            if (!orderBehavior) {
                                let prior = demand.suppliedBy;
                                if (prior != null && prior.orderingState == OrderingState.Ordered && prior.order >= behavior.order) {
                                    orderBehavior = true;
                                }
                            }
                        }
                    }
                    let newDemands = null;
                    let orderingDemands = null;
                    for (let linkable of allUntrackedDemands) {
                        if (newDemands == null) {
                            newDemands = new Set();
                        }
                        newDemands.add(linkable.resource);
                        if (linkable.type == LinkType.order) {
                            if (orderingDemands == null) {
                                orderingDemands = new Set();
                            }
                            orderingDemands.add(linkable.resource);
                        }
                    }
                    behavior.demands = newDemands;
                    behavior.orderingDemands = orderingDemands;
                    if (orderBehavior) {
                        if (behavior.orderingState != OrderingState.NeedsOrdering) {
                            behavior.orderingState = OrderingState.NeedsOrdering;
                            this.needsOrdering.push(behavior);
                        }
                    }
                    if (needsRunning) {
                        this.activateBehavior(behavior, sequence);
                    }
                }
                this.modifiedDemandBehaviors.length = 0;
            }
        }
        orderBehaviors() {
            // find all behaviors that need ordering and their
            // subsequents and mark them all as needing ordering
            if (this.needsOrdering.length == 0) {
                return;
            }
            let localNeedsOrdering = [];
            // dfs forward on each to find all that need ordering
            let x = 0;
            while (x < this.needsOrdering.length) {
                let behavior = this.needsOrdering[x];
                if (behavior.orderingState == OrderingState.NeedsOrdering) {
                    behavior.orderingState = OrderingState.Clearing;
                    localNeedsOrdering.push(behavior);
                    if (behavior.supplies) {
                        for (let supply of behavior.supplies) {
                            for (let subsequent of supply.subsequents) {
                                if (subsequent.orderingState == OrderingState.Ordered) {
                                    subsequent.orderingState = OrderingState.NeedsOrdering;
                                    this.needsOrdering.push(subsequent);
                                }
                            }
                        }
                    }
                }
                x++;
            }
            this.needsOrdering.length = 0;
            let needsReheap = { value: false }; // this allows out parameter
            for (let behavior of localNeedsOrdering) {
                this.sortDFS(behavior, needsReheap);
            }
            if (needsReheap.value) {
                this.activatedBehaviors.unsort();
            }
        }
        sortDFS(behavior, needsReheap) {
            if (behavior.orderingState == OrderingState.Ordering) {
                let err = new Error("Behavior dependency cycle detected.");
                err.currentBehavior = behavior;
                err.cycle = this.debugCycleForBehavior(behavior);
                throw err;
            }
            if (behavior.orderingState == OrderingState.Clearing) {
                behavior.orderingState = OrderingState.Ordering;
                let order = 0;
                if (behavior.demands != null) {
                    for (let demand of behavior.demands) {
                        let prior = demand.suppliedBy;
                        if (prior != null) {
                            if (prior.orderingState != OrderingState.Ordered) {
                                this.sortDFS(prior, needsReheap);
                            }
                            order = Math.max(order, prior.order + 1);
                        }
                    }
                }
                behavior.orderingState = OrderingState.Ordered;
                if (order != behavior.order) {
                    behavior.order = order;
                    needsReheap.value = true;
                }
            }
        }
        debugCycleForBehavior(behavior) {
            let stack = [];
            let output = [];
            if (this.cycleDFS(behavior, behavior, stack)) {
                output = stack;
            }
            return output;
        }
        cycleDFS(currentBehavior, target, stack) {
            if (currentBehavior.demands != null) {
                for (let r of currentBehavior.demands) {
                    let b = r.suppliedBy;
                    if (b != null) {
                        stack.push(r);
                        if (b == target) {
                            return true;
                        }
                        if (this.cycleDFS(b, target, stack)) {
                            return true;
                        }
                        stack.pop();
                    }
                }
            }
            return false;
        }
        addBehavior(behavior) {
            this.untrackedBehaviors.push(behavior);
        }
        updateDemands(behavior, newDemands) {
            if (behavior.extent.addedToGraphWhen == null) {
                let err = new Error("Behavior must belong to graph before updating demands.");
                err.behavior = behavior;
                throw err;
            }
            else if (this.currentEvent == null) {
                let err = new Error("Demands can only be updated during an event.");
                err.behavior = behavior;
                throw err;
            }
            behavior.untrackedDynamicDemands = newDemands;
            this.modifiedDemandBehaviors.push(behavior);
        }
        updateSupplies(behavior, newSupplies) {
            if (behavior.extent.addedToGraphWhen == null) {
                let err = new Error("Behavior must belong to graph before updating supplies.");
                err.behavior = behavior;
                throw err;
            }
            else if (this.currentEvent == null) {
                let err = new Error("Supplies can only be updated during an event.");
                err.behavior = behavior;
                throw err;
            }
            behavior.untrackedDynamicSupplies = newSupplies;
            this.modifiedSupplyBehaviors.push(behavior);
        }
        removeBehavior(behavior, sequence) {
            // If we demand a foreign resource then we should be
            // removed from its list of subsequents
            if (behavior.demands != null) {
                let removed = false;
                for (let demand of behavior.demands) {
                    if (demand.extent != behavior.extent) {
                        demand.subsequents.delete(behavior);
                        removed = true;
                    }
                }
                // and remove foreign demands
                // its faster to erase the whole list than pick out the foreign ones
                if (removed) {
                    behavior.demands.clear();
                }
            }
            // any foreign resources should no longer be supplied by this behavior
            if (behavior.supplies != null) {
                let removed = false;
                for (let supply of behavior.supplies) {
                    if (supply.extent != behavior.extent) {
                        supply.suppliedBy = null;
                        removed = true;
                    }
                }
                // and clear out those foreign supplies
                // its faster to clear whole list than pick out individual foreign ones
                if (removed) {
                    behavior.supplies.clear();
                }
            }
            behavior.removedWhen = sequence;
        }
        addExtent(extent) {
            if (extent.addedToGraphWhen != null) {
                let err = new Error("Extent already belongs to a graph.");
                err.extent = extent;
                err.graph = this;
                throw err;
            }
            if (this.currentEvent == null) {
                let err = new Error("Extents can only be added during an event.");
                err.extent = extent;
                throw err;
            }
            if (this.validateLifetimes) {
                if (extent.lifetime != null) {
                    if (extent.lifetime.addedToGraphWhen == null) {
                        // first extent in lifetime being added
                        extent.lifetime.addedToGraphWhen = this.currentEvent.sequence;
                    }
                    if (extent.lifetime.parent != null) {
                        if (extent.lifetime.parent.addedToGraphWhen == null) {
                            let err = new Error("Extent with child lifetime must be added after parent.");
                            err.extent = extent;
                            throw err;
                        }
                    }
                }
            }
            extent.addedToGraphWhen = this.currentEvent.sequence;
            this.extentsAdded.push(extent);
            // this casting below is a hack to get at the private method so we can skip integrity checks which
            // allow us to update addedToGraph from inside whatever behavior or action it is added
            extent.addedToGraph._updateForce(true);
            for (let behavior of extent.behaviors) {
                this.addBehavior(behavior);
            }
        }
        removeExtent(extent) {
            if (this.currentEvent == null) {
                let err = new Error("Extents can only be removed during an event.");
                err.extent = extent;
                throw err;
            }
            else {
                this.extentsRemoved.push(extent);
                for (let behavior of extent.behaviors) {
                    this.removeBehavior(behavior, this.currentEvent.sequence);
                }
                extent.unsubscribeAll();
                extent.addedToGraphWhen = null;
            }
        }
    }
    var EventLoopPhase;
    (function (EventLoopPhase) {
        EventLoopPhase[EventLoopPhase["queued"] = 0] = "queued";
        EventLoopPhase[EventLoopPhase["action"] = 1] = "action";
        EventLoopPhase[EventLoopPhase["updates"] = 2] = "updates";
        EventLoopPhase[EventLoopPhase["sideEffects"] = 3] = "sideEffects";
    })(EventLoopPhase || (EventLoopPhase = {}));
    class EventLoopState {
        action;
        actionUpdates;
        currentSideEffect = null;
        phase;
        constructor(action) {
            this.action = action;
            this.phase = EventLoopPhase.queued;
            this.actionUpdates = [];
        }
    }

    //
    class Behavior {
        demands;
        orderingDemands;
        supplies;
        block;
        enqueuedWhen = null;
        removedWhen = null;
        extent;
        orderingState = OrderingState.Untracked;
        order = 0;
        untrackedDemands;
        untrackedDynamicDemands;
        untrackedSupplies;
        untrackedDynamicSupplies;
        constructor(extent, demands, supplies, block) {
            this.extent = extent;
            extent.addBehavior(this);
            this.demands = null;
            this.orderingDemands = null;
            this.supplies = null;
            this.block = block;
            this.untrackedDemands = demands;
            this.untrackedDynamicDemands = null;
            this.untrackedSupplies = supplies;
            this.untrackedDynamicSupplies = null;
        }
        toString() {
            let name = "Behavior (" + this.order + ")";
            if (this.enqueuedWhen != null) {
                name = name + " : " + this.enqueuedWhen;
            }
            if (this.supplies != null && this.supplies?.size > 0) {
                name = name + " \n Supplies:";
                this.supplies?.forEach(item => {
                    name = name + "\n  " + item.toString();
                });
            }
            if (this.demands != null && this.demands?.size > 0) {
                name = name + " \n Demands:";
                this.demands?.forEach(item => {
                    name = name + "\n  " + item.toString();
                });
            }
            return name;
        }
        setDynamicDemands(newDemands) {
            this.extent.graph.updateDemands(this, newDemands?.filter(item => item != undefined));
        }
        setDynamicSupplies(newSupplies) {
            this.extent.graph.updateSupplies(this, newSupplies?.filter(item => item !== undefined));
        }
    }
    class BehaviorBuilder {
        extent;
        untrackedDemands = null;
        untrackedSupplies = null;
        dynamicDemandSwitches = null;
        dynamicDemandLinks = null;
        dynamicDemandRelinkingOrder = exports.RelinkingOrder.relinkingOrderPrior;
        dynamicSupplySwitches = null;
        dynamicSupplyLinks = null;
        dynamicSupplyRelinkingOrder = exports.RelinkingOrder.relinkingOrderPrior;
        constructor(extent) {
            this.extent = extent;
        }
        demands(...demands) {
            this.untrackedDemands = demands;
            return this;
        }
        supplies(...supplies) {
            this.untrackedSupplies = supplies;
            return this;
        }
        dynamicDemands(switches, links, relinkingOrder) {
            this.dynamicDemandSwitches = switches;
            this.dynamicDemandLinks = links;
            if (relinkingOrder != undefined) {
                this.dynamicDemandRelinkingOrder = relinkingOrder;
            }
            return this;
        }
        dynamicSupplies(switches, links, relinkingOrder) {
            this.dynamicSupplySwitches = switches;
            this.dynamicSupplyLinks = links;
            if (relinkingOrder != undefined) {
                this.dynamicSupplyRelinkingOrder = relinkingOrder;
            }
            return this;
        }
        runs(block) {
            let hasDynamicDemands = this.dynamicDemandSwitches != null;
            if (this.untrackedDemands == null) {
                this.untrackedDemands = [];
            }
            if (this.untrackedSupplies == null) {
                this.untrackedSupplies = [];
            }
            let dynamicDemandResource;
            if (hasDynamicDemands) {
                dynamicDemandResource = this.extent.resource('(BG Dynamic Demand Resource)');
                if (this.dynamicDemandRelinkingOrder == exports.RelinkingOrder.relinkingOrderPrior) {
                    this.untrackedDemands.push(dynamicDemandResource);
                }
                else {
                    this.untrackedSupplies.push(dynamicDemandResource);
                }
            }
            let hasDynamicSupplies = this.dynamicSupplySwitches != null;
            let dynamicSupplyResource;
            if (hasDynamicSupplies) {
                dynamicSupplyResource = this.extent.resource('(BG Dynamic Supply Resource)');
                if (this.dynamicSupplyRelinkingOrder == exports.RelinkingOrder.relinkingOrderPrior) {
                    this.untrackedDemands.push(dynamicSupplyResource);
                }
                else {
                    this.untrackedSupplies.push(dynamicSupplyResource);
                }
            }
            let mainBehavior = new Behavior(this.extent, this.untrackedDemands, this.untrackedSupplies, block);
            if (hasDynamicDemands) {
                let supplies = [];
                let demands = this.dynamicDemandSwitches;
                if (this.dynamicDemandRelinkingOrder == exports.RelinkingOrder.relinkingOrderPrior) {
                    supplies.push(dynamicDemandResource);
                }
                else {
                    if (demands == null) {
                        demands = [];
                    }
                    demands.push(dynamicDemandResource);
                }
                new Behavior(this.extent, demands, supplies, ((extent) => {
                    let demandLinks = this.dynamicDemandLinks(extent);
                    mainBehavior.setDynamicDemands(demandLinks);
                }));
            }
            if (hasDynamicSupplies) {
                let supplies = [];
                let demands = this.dynamicSupplySwitches;
                if (this.dynamicSupplyRelinkingOrder == exports.RelinkingOrder.relinkingOrderPrior) {
                    supplies.push(dynamicSupplyResource);
                }
                else {
                    if (demands == null) {
                        demands = [];
                    }
                    demands.push(dynamicSupplyResource);
                }
                new Behavior(this.extent, demands, supplies, ((extent) => {
                    let supplyLinks = this.dynamicSupplyLinks(extent);
                    mainBehavior.setDynamicSupplies(supplyLinks);
                }));
            }
            return mainBehavior;
        }
    }

    //
    exports.ExtentRemoveStrategy = void 0;
    (function (ExtentRemoveStrategy) {
        ExtentRemoveStrategy[ExtentRemoveStrategy["extentOnly"] = 0] = "extentOnly";
        ExtentRemoveStrategy[ExtentRemoveStrategy["containedLifetimes"] = 1] = "containedLifetimes";
    })(exports.ExtentRemoveStrategy || (exports.ExtentRemoveStrategy = {}));
    class ExtentLifetime {
        addedToGraphWhen = null;
        extents = new Set();
        children = null;
        parent = null;
        constructor(extent) {
            this.extents.add(extent);
            if (extent.addedToGraphWhen != null) {
                this.addedToGraphWhen = extent.addedToGraphWhen;
            }
        }
        unify(extent) {
            if (extent.addedToGraphWhen != null) {
                let err = new Error("Same lifetime relationship must be established before adding any extent to graph.");
                err.extent = extent;
                throw err;
            }
            if (extent.lifetime != null) {
                // merge existing lifetimes and children into one lifetime heirarchy
                // move children first
                if (extent.lifetime.children != null) {
                    for (let child of extent.lifetime.children) {
                        this.addChildLifetime(child);
                    }
                }
                // then make any extents in other lifetime part of this one
                for (let ext of extent.lifetime.extents) {
                    ext.lifetime = this;
                    this.extents.add(ext);
                }
            }
            else {
                extent.lifetime = this;
                this.extents.add(extent);
            }
        }
        addChild(extent) {
            if (extent.lifetime == null) {
                extent.lifetime = new ExtentLifetime(extent);
            }
            this.addChildLifetime(extent.lifetime);
        }
        addChildLifetime(lifetime) {
            let myLifetime = this;
            while (myLifetime != null) {
                if (myLifetime === lifetime) {
                    let err = new Error("Child lifetime cannot be a transitive parent.");
                    err.extent = lifetime;
                    throw err;
                }
                myLifetime = myLifetime.parent;
            }
            lifetime.parent = this;
            if (this.children == null) {
                this.children = new Set();
            }
            this.children.add(lifetime);
        }
        hasCompatibleLifetime(lifetime) {
            if (this === lifetime) {
                // unified
                return true;
            }
            else if (lifetime != null) {
                // parents
                if (this.parent != null) {
                    return this.parent.hasCompatibleLifetime(lifetime);
                }
            }
            return false;
        }
        getAllContainedExtents() {
            let extents = [];
            for (let ext of this.extents) {
                extents.push(ext);
            }
            if (this.children != null) {
                for (let childLifetime of this.children) {
                    extents.push(...childLifetime.getAllContainedExtents());
                }
            }
            return extents;
        }
        getAllContainingExtents() {
            let extents = [];
            for (let ext of this.extents) {
                extents.push(ext);
            }
            if (this.parent != null) {
                extents.push(...this.parent.getAllContainingExtents());
            }
            return extents;
        }
    }
    class Extent {
        debugConstructorName;
        debugName;
        behaviors = [];
        resources = [];
        graph;
        addedToGraphWhen = null;
        addedToGraph;
        lifetime = null;
        unsubscribes = new Set();
        static removeContainedLifetimes = exports.ExtentRemoveStrategy.containedLifetimes;
        static relinkingOrderSubsequent = exports.RelinkingOrder.relinkingOrderSubsequent;
        constructor(graph) {
            if (graph === null || graph === undefined) {
                let err = new Error("Extent must be initialized with an instance of Graph");
                err.extent = this;
                throw err;
            }
            this.debugConstructorName = this.constructor.name;
            this.graph = graph;
            this.addedToGraph = new State(this, false);
        }
        debugHere() {
            return this.graph.debugHere();
        }
        unifyLifetime(extent) {
            if (this.lifetime == null) {
                this.lifetime = new ExtentLifetime(this);
            }
            this.lifetime.unify(extent);
        }
        addChildLifetime(extent) {
            if (this.lifetime == null) {
                this.lifetime = new ExtentLifetime(this);
            }
            this.lifetime.addChild(extent);
        }
        hasCompatibleLifetime(extent) {
            if (this === extent) {
                return true;
            }
            else if (this.lifetime != null) {
                return (this.lifetime.hasCompatibleLifetime(extent.lifetime));
            }
            else {
                return false;
            }
        }
        addBehavior(behavior) {
            this.behaviors.push(behavior);
        }
        addResource(resource) {
            this.resources.push(resource);
        }
        addToGraphWithAction(debugName) {
            this.graph.action(() => {
                this.addToGraph();
            }, debugName);
        }
        addToGraph() {
            if (this.graph.currentEvent != null) {
                this.nameResources();
                this.graph.addExtent(this);
            }
            else {
                let err = new Error("addToGraph must be called within an event.");
                err.extent = this;
                throw err;
            }
        }
        removeFromGraphWithAction(strategy, debugName) {
            this.action(() => {
                this.removeFromGraph(strategy);
            }, debugName);
        }
        removeFromGraph(strategy) {
            let graph = this.graph;
            if (graph.currentEvent != null) {
                if (this.addedToGraphWhen != null) {
                    if (strategy == exports.ExtentRemoveStrategy.extentOnly || strategy === undefined || this.lifetime === null) {
                        graph.removeExtent(this);
                    }
                    else {
                        for (let ext of this.lifetime.getAllContainedExtents()) {
                            graph.removeExtent(ext);
                        }
                    }
                }
            }
            else {
                let err = new Error("removeFromGraph must be called within an event.");
                err.extent = this;
                throw err;
            }
        }
        subscribeToJustUpdated(resources, callback) {
            let unsubscribe = this.graph._subscribeToJustUpdated(resources, { extent: this, callback: callback });
            this.unsubscribes.add(unsubscribe);
            return unsubscribe;
        }
        unsubscribeAll() {
            for (let unsubscribe of this.unsubscribes) {
                unsubscribe();
            }
            this.unsubscribes.clear();
        }
        nameResources() {
            // automatically add any behaviors and resources that are contained
            // by this Extent object and name them with corresponding keys
            for (let key in this) {
                let object = this[key];
                if (object && object['isResource'] !== undefined) {
                    if (object.debugName == null) {
                        object.debugName = key;
                    }
                }
            }
        }
        behavior() {
            let b = new BehaviorBuilder(this);
            return b;
        }
        resource(name) {
            return new Resource(this, name);
        }
        moment(name) {
            return new Moment(this, name);
        }
        state(initialState, name) {
            return new State(this, initialState, name);
        }
        sideEffect(block, debugName) {
            // This requires a cast because we know the extent won't be null at runtime because this side effect
            // was created with one
            this.graph.sideEffectHelper({
                debugName: debugName,
                block: block,
                extent: this,
                behavior: this.graph.currentBehavior
            });
        }
        async actionAsync(action, debugName) {
            return this.graph.actionAsyncHelper({
                block: action,
                debugName: debugName,
                extent: this,
                resolve: null
            });
        }
        action(action, debugName) {
            this.graph.actionHelper({
                block: action,
                debugName: debugName,
                extent: this,
                resolve: null
            });
        }
    }

    exports.Behavior = Behavior;
    exports.Extent = Extent;
    exports.Graph = Graph;
    exports.GraphEvent = GraphEvent;
    exports.Moment = Moment;
    exports.Resource = Resource;
    exports.State = State;

    Object.defineProperty(exports, '__esModule', { value: true });

    return exports;

})({});
