"use strict";
//
//  Copyright Yahoo 2021
//
Object.defineProperty(exports, "__esModule", { value: true });
exports.BehaviorBuilder = exports.Behavior = void 0;
const common_1 = require("./common");
class Behavior {
    constructor(extent, demands, supplies, block) {
        this.enqueuedWhen = null;
        this.removedWhen = null;
        this.orderingState = common_1.OrderingState.Untracked;
        this.order = 0;
        this.extent = extent;
        extent.addBehavior(this);
        this.demands = null;
        this.orderingDemands = null;
        this.supplies = null;
        this.block = block;
        this.untrackedDemands = demands;
        this.untrackedDynamicDemands = null;
        this.untrackedSupplies = supplies;
        this.untrackedDynamicSupplies = null;
    }
    toString() {
        var _a, _b, _c, _d;
        let name = "Behavior (" + this.order + ")";
        if (this.enqueuedWhen != null) {
            name = name + " : " + this.enqueuedWhen;
        }
        if (this.supplies != null && ((_a = this.supplies) === null || _a === void 0 ? void 0 : _a.size) > 0) {
            name = name + " \n Supplies:";
            (_b = this.supplies) === null || _b === void 0 ? void 0 : _b.forEach(item => {
                name = name + "\n  " + item.toString();
            });
        }
        if (this.demands != null && ((_c = this.demands) === null || _c === void 0 ? void 0 : _c.size) > 0) {
            name = name + " \n Demands:";
            (_d = this.demands) === null || _d === void 0 ? void 0 : _d.forEach(item => {
                name = name + "\n  " + item.toString();
            });
        }
        return name;
    }
    setDynamicDemands(newDemands) {
        this.extent.graph.updateDemands(this, newDemands === null || newDemands === void 0 ? void 0 : newDemands.filter(item => item != undefined));
    }
    setDynamicSupplies(newSupplies) {
        this.extent.graph.updateSupplies(this, newSupplies === null || newSupplies === void 0 ? void 0 : newSupplies.filter(item => item !== undefined));
    }
}
exports.Behavior = Behavior;
class BehaviorBuilder {
    constructor(extent) {
        this.untrackedDemands = null;
        this.untrackedSupplies = null;
        this.dynamicDemandSwitches = null;
        this.dynamicDemandLinks = null;
        this.dynamicDemandRelinkingOrder = common_1.RelinkingOrder.relinkingOrderPrior;
        this.dynamicSupplySwitches = null;
        this.dynamicSupplyLinks = null;
        this.dynamicSupplyRelinkingOrder = common_1.RelinkingOrder.relinkingOrderPrior;
        this.extent = extent;
    }
    demands(...demands) {
        this.untrackedDemands = demands;
        return this;
    }
    supplies(...supplies) {
        this.untrackedSupplies = supplies;
        return this;
    }
    dynamicDemands(switches, links, relinkingOrder) {
        this.dynamicDemandSwitches = switches;
        this.dynamicDemandLinks = links;
        if (relinkingOrder != undefined) {
            this.dynamicDemandRelinkingOrder = relinkingOrder;
        }
        return this;
    }
    dynamicSupplies(switches, links, relinkingOrder) {
        this.dynamicSupplySwitches = switches;
        this.dynamicSupplyLinks = links;
        if (relinkingOrder != undefined) {
            this.dynamicSupplyRelinkingOrder = relinkingOrder;
        }
        return this;
    }
    runs(block) {
        let hasDynamicDemands = this.dynamicDemandSwitches != null;
        if (this.untrackedDemands == null) {
            this.untrackedDemands = [];
        }
        if (this.untrackedSupplies == null) {
            this.untrackedSupplies = [];
        }
        let dynamicDemandResource;
        if (hasDynamicDemands) {
            dynamicDemandResource = this.extent.resource('(BG Dynamic Demand Resource)');
            if (this.dynamicDemandRelinkingOrder == common_1.RelinkingOrder.relinkingOrderPrior) {
                this.untrackedDemands.push(dynamicDemandResource);
            }
            else {
                this.untrackedSupplies.push(dynamicDemandResource);
            }
        }
        let hasDynamicSupplies = this.dynamicSupplySwitches != null;
        let dynamicSupplyResource;
        if (hasDynamicSupplies) {
            dynamicSupplyResource = this.extent.resource('(BG Dynamic Supply Resource)');
            if (this.dynamicSupplyRelinkingOrder == common_1.RelinkingOrder.relinkingOrderPrior) {
                this.untrackedDemands.push(dynamicSupplyResource);
            }
            else {
                this.untrackedSupplies.push(dynamicSupplyResource);
            }
        }
        let mainBehavior = new Behavior(this.extent, this.untrackedDemands, this.untrackedSupplies, block);
        if (hasDynamicDemands) {
            let supplies = [];
            let demands = this.dynamicDemandSwitches;
            if (this.dynamicDemandRelinkingOrder == common_1.RelinkingOrder.relinkingOrderPrior) {
                supplies.push(dynamicDemandResource);
            }
            else {
                if (demands == null) {
                    demands = [];
                }
                demands.push(dynamicDemandResource);
            }
            new Behavior(this.extent, demands, supplies, ((extent) => {
                let demandLinks = this.dynamicDemandLinks(extent);
                mainBehavior.setDynamicDemands(demandLinks);
            }));
        }
        if (hasDynamicSupplies) {
            let supplies = [];
            let demands = this.dynamicSupplySwitches;
            if (this.dynamicSupplyRelinkingOrder == common_1.RelinkingOrder.relinkingOrderPrior) {
                supplies.push(dynamicSupplyResource);
            }
            else {
                if (demands == null) {
                    demands = [];
                }
                demands.push(dynamicSupplyResource);
            }
            new Behavior(this.extent, demands, supplies, ((extent) => {
                let supplyLinks = this.dynamicSupplyLinks(extent);
                mainBehavior.setDynamicSupplies(supplyLinks);
            }));
        }
        return mainBehavior;
    }
}
exports.BehaviorBuilder = BehaviorBuilder;
//# sourceMappingURL=behavior.js.map