"use strict";
//
//  Copyright Yahoo 2021
//
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Extent = exports.ExtentRemoveStrategy = void 0;
const behavior_js_1 = require("./behavior.js");
const resource_js_1 = require("./resource.js");
const common_js_1 = require("./common.js");
var ExtentRemoveStrategy;
(function (ExtentRemoveStrategy) {
    ExtentRemoveStrategy[ExtentRemoveStrategy["extentOnly"] = 0] = "extentOnly";
    ExtentRemoveStrategy[ExtentRemoveStrategy["containedLifetimes"] = 1] = "containedLifetimes";
})(ExtentRemoveStrategy = exports.ExtentRemoveStrategy || (exports.ExtentRemoveStrategy = {}));
class ExtentLifetime {
    constructor(extent) {
        this.addedToGraphWhen = null;
        this.extents = new Set();
        this.children = null;
        this.parent = null;
        this.extents.add(extent);
        if (extent.addedToGraphWhen != null) {
            this.addedToGraphWhen = extent.addedToGraphWhen;
        }
    }
    unify(extent) {
        if (extent.addedToGraphWhen != null) {
            let err = new Error("Same lifetime relationship must be established before adding any extent to graph.");
            err.extent = extent;
            throw err;
        }
        if (extent.lifetime != null) {
            // merge existing lifetimes and children into one lifetime heirarchy
            // move children first
            if (extent.lifetime.children != null) {
                for (let child of extent.lifetime.children) {
                    this.addChildLifetime(child);
                }
            }
            // then make any extents in other lifetime part of this one
            for (let ext of extent.lifetime.extents) {
                ext.lifetime = this;
                this.extents.add(ext);
            }
        }
        else {
            extent.lifetime = this;
            this.extents.add(extent);
        }
    }
    addChild(extent) {
        if (extent.lifetime == null) {
            extent.lifetime = new ExtentLifetime(extent);
        }
        this.addChildLifetime(extent.lifetime);
    }
    addChildLifetime(lifetime) {
        let myLifetime = this;
        while (myLifetime != null) {
            if (myLifetime === lifetime) {
                let err = new Error("Child lifetime cannot be a transitive parent.");
                err.extent = lifetime;
                throw err;
            }
            myLifetime = myLifetime.parent;
        }
        lifetime.parent = this;
        if (this.children == null) {
            this.children = new Set();
        }
        this.children.add(lifetime);
    }
    hasCompatibleLifetime(lifetime) {
        if (this === lifetime) {
            // unified
            return true;
        }
        else if (lifetime != null) {
            // parents
            if (this.parent != null) {
                return this.parent.hasCompatibleLifetime(lifetime);
            }
        }
        return false;
    }
    getAllContainedExtents() {
        let extents = [];
        for (let ext of this.extents) {
            extents.push(ext);
        }
        if (this.children != null) {
            for (let childLifetime of this.children) {
                extents.push(...childLifetime.getAllContainedExtents());
            }
        }
        return extents;
    }
    getAllContainingExtents() {
        let extents = [];
        for (let ext of this.extents) {
            extents.push(ext);
        }
        if (this.parent != null) {
            extents.push(...this.parent.getAllContainingExtents());
        }
        return extents;
    }
}
class Extent {
    constructor(graph) {
        this.behaviors = [];
        this.resources = [];
        this.addedToGraphWhen = null;
        this.lifetime = null;
        this.unsubscribes = new Set();
        if (graph === null || graph === undefined) {
            let err = new Error("Extent must be initialized with an instance of Graph");
            err.extent = this;
            throw err;
        }
        this.debugConstructorName = this.constructor.name;
        this.graph = graph;
        this.addedToGraph = new resource_js_1.State(this, false);
    }
    debugHere() {
        return this.graph.debugHere();
    }
    unifyLifetime(extent) {
        if (this.lifetime == null) {
            this.lifetime = new ExtentLifetime(this);
        }
        this.lifetime.unify(extent);
    }
    addChildLifetime(extent) {
        if (this.lifetime == null) {
            this.lifetime = new ExtentLifetime(this);
        }
        this.lifetime.addChild(extent);
    }
    hasCompatibleLifetime(extent) {
        if (this === extent) {
            return true;
        }
        else if (this.lifetime != null) {
            return (this.lifetime.hasCompatibleLifetime(extent.lifetime));
        }
        else {
            return false;
        }
    }
    addBehavior(behavior) {
        this.behaviors.push(behavior);
    }
    addResource(resource) {
        this.resources.push(resource);
    }
    addToGraphWithAction(debugName) {
        this.graph.action(() => {
            this.addToGraph();
        }, debugName);
    }
    addToGraph() {
        if (this.graph.currentEvent != null) {
            this.nameResources();
            this.graph.addExtent(this);
        }
        else {
            let err = new Error("addToGraph must be called within an event.");
            err.extent = this;
            throw err;
        }
    }
    removeFromGraphWithAction(strategy, debugName) {
        this.action(() => {
            this.removeFromGraph(strategy);
        }, debugName);
    }
    removeFromGraph(strategy) {
        let graph = this.graph;
        if (graph.currentEvent != null) {
            if (this.addedToGraphWhen != null) {
                if (strategy == ExtentRemoveStrategy.extentOnly || strategy === undefined || this.lifetime === null) {
                    graph.removeExtent(this);
                }
                else {
                    for (let ext of this.lifetime.getAllContainedExtents()) {
                        graph.removeExtent(ext);
                    }
                }
            }
        }
        else {
            let err = new Error("removeFromGraph must be called within an event.");
            err.extent = this;
            throw err;
        }
    }
    subscribeToJustUpdated(resources, callback) {
        let unsubscribe = this.graph._subscribeToJustUpdated(resources, { extent: this, callback: callback });
        this.unsubscribes.add(unsubscribe);
        return unsubscribe;
    }
    unsubscribeAll() {
        for (let unsubscribe of this.unsubscribes) {
            unsubscribe();
        }
        this.unsubscribes.clear();
    }
    nameResources() {
        // automatically add any behaviors and resources that are contained
        // by this Extent object and name them with corresponding keys
        for (let key in this) {
            let object = this[key];
            if (object && object['isResource'] !== undefined) {
                if (object.debugName == null) {
                    object.debugName = key;
                }
            }
        }
    }
    behavior() {
        let b = new behavior_js_1.BehaviorBuilder(this);
        return b;
    }
    resource(name) {
        return new resource_js_1.Resource(this, name);
    }
    moment(name) {
        return new resource_js_1.Moment(this, name);
    }
    state(initialState, name) {
        return new resource_js_1.State(this, initialState, name);
    }
    sideEffect(block, debugName) {
        // This requires a cast because we know the extent won't be null at runtime because this side effect
        // was created with one
        this.graph.sideEffectHelper({
            debugName: debugName,
            block: block,
            extent: this,
            behavior: this.graph.currentBehavior
        });
    }
    actionAsync(action, debugName) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.graph.actionAsyncHelper({
                block: action,
                debugName: debugName,
                extent: this,
                resolve: null
            });
        });
    }
    action(action, debugName) {
        this.graph.actionHelper({
            block: action,
            debugName: debugName,
            extent: this,
            resolve: null
        });
    }
}
exports.Extent = Extent;
Extent.removeContainedLifetimes = ExtentRemoveStrategy.containedLifetimes;
Extent.relinkingOrderSubsequent = common_js_1.RelinkingOrder.relinkingOrderSubsequent;
//# sourceMappingURL=extent.js.map