//
//  Copyright Yahoo 2021
//
import { OrderingState, RelinkingOrder } from "./common";
export class Behavior {
    demands;
    orderingDemands;
    supplies;
    block;
    enqueuedWhen = null;
    removedWhen = null;
    extent;
    orderingState = OrderingState.Untracked;
    order = 0;
    untrackedDemands;
    untrackedDynamicDemands;
    untrackedSupplies;
    untrackedDynamicSupplies;
    constructor(extent, demands, supplies, block) {
        this.extent = extent;
        extent.addBehavior(this);
        this.demands = null;
        this.orderingDemands = null;
        this.supplies = null;
        this.block = block;
        this.untrackedDemands = demands;
        this.untrackedDynamicDemands = null;
        this.untrackedSupplies = supplies;
        this.untrackedDynamicSupplies = null;
    }
    toString() {
        let name = "Behavior (" + this.order + ")";
        if (this.enqueuedWhen != null) {
            name = name + " : " + this.enqueuedWhen;
        }
        if (this.supplies != null && this.supplies?.size > 0) {
            name = name + " \n Supplies:";
            this.supplies?.forEach(item => {
                name = name + "\n  " + item.toString();
            });
        }
        if (this.demands != null && this.demands?.size > 0) {
            name = name + " \n Demands:";
            this.demands?.forEach(item => {
                name = name + "\n  " + item.toString();
            });
        }
        return name;
    }
    setDynamicDemands(newDemands) {
        this.extent.graph.updateDemands(this, newDemands?.filter(item => item != undefined));
    }
    setDynamicSupplies(newSupplies) {
        this.extent.graph.updateSupplies(this, newSupplies?.filter(item => item !== undefined));
    }
}
export class BehaviorBuilder {
    extent;
    untrackedDemands = null;
    untrackedSupplies = null;
    dynamicDemandSwitches = null;
    dynamicDemandLinks = null;
    dynamicDemandRelinkingOrder = RelinkingOrder.relinkingOrderPrior;
    dynamicSupplySwitches = null;
    dynamicSupplyLinks = null;
    dynamicSupplyRelinkingOrder = RelinkingOrder.relinkingOrderPrior;
    constructor(extent) {
        this.extent = extent;
    }
    demands(...demands) {
        this.untrackedDemands = demands;
        return this;
    }
    supplies(...supplies) {
        this.untrackedSupplies = supplies;
        return this;
    }
    dynamicDemands(switches, links, relinkingOrder) {
        this.dynamicDemandSwitches = switches;
        this.dynamicDemandLinks = links;
        if (relinkingOrder != undefined) {
            this.dynamicDemandRelinkingOrder = relinkingOrder;
        }
        return this;
    }
    dynamicSupplies(switches, links, relinkingOrder) {
        this.dynamicSupplySwitches = switches;
        this.dynamicSupplyLinks = links;
        if (relinkingOrder != undefined) {
            this.dynamicSupplyRelinkingOrder = relinkingOrder;
        }
        return this;
    }
    runs(block) {
        let hasDynamicDemands = this.dynamicDemandSwitches != null;
        if (this.untrackedDemands == null) {
            this.untrackedDemands = [];
        }
        if (this.untrackedSupplies == null) {
            this.untrackedSupplies = [];
        }
        let dynamicDemandResource;
        if (hasDynamicDemands) {
            dynamicDemandResource = this.extent.resource('(BG Dynamic Demand Resource)');
            if (this.dynamicDemandRelinkingOrder == RelinkingOrder.relinkingOrderPrior) {
                this.untrackedDemands.push(dynamicDemandResource);
            }
            else {
                this.untrackedSupplies.push(dynamicDemandResource);
            }
        }
        let hasDynamicSupplies = this.dynamicSupplySwitches != null;
        let dynamicSupplyResource;
        if (hasDynamicSupplies) {
            dynamicSupplyResource = this.extent.resource('(BG Dynamic Supply Resource)');
            if (this.dynamicSupplyRelinkingOrder == RelinkingOrder.relinkingOrderPrior) {
                this.untrackedDemands.push(dynamicSupplyResource);
            }
            else {
                this.untrackedSupplies.push(dynamicSupplyResource);
            }
        }
        let mainBehavior = new Behavior(this.extent, this.untrackedDemands, this.untrackedSupplies, block);
        if (hasDynamicDemands) {
            let supplies = [];
            let demands = this.dynamicDemandSwitches;
            if (this.dynamicDemandRelinkingOrder == RelinkingOrder.relinkingOrderPrior) {
                supplies.push(dynamicDemandResource);
            }
            else {
                if (demands == null) {
                    demands = [];
                }
                demands.push(dynamicDemandResource);
            }
            new Behavior(this.extent, demands, supplies, ((extent) => {
                let demandLinks = this.dynamicDemandLinks(extent);
                mainBehavior.setDynamicDemands(demandLinks);
            }));
        }
        if (hasDynamicSupplies) {
            let supplies = [];
            let demands = this.dynamicSupplySwitches;
            if (this.dynamicSupplyRelinkingOrder == RelinkingOrder.relinkingOrderPrior) {
                supplies.push(dynamicSupplyResource);
            }
            else {
                if (demands == null) {
                    demands = [];
                }
                demands.push(dynamicSupplyResource);
            }
            new Behavior(this.extent, demands, supplies, ((extent) => {
                let supplyLinks = this.dynamicSupplyLinks(extent);
                mainBehavior.setDynamicSupplies(supplyLinks);
            }));
        }
        return mainBehavior;
    }
}
//# sourceMappingURL=behavior.js.map