//
//  Copyright Yahoo 2021
//
import { GraphEvent } from "./common";
export var LinkType;
(function (LinkType) {
    LinkType[LinkType["reactive"] = 0] = "reactive";
    LinkType[LinkType["order"] = 1] = "order";
})(LinkType || (LinkType = {}));
export class Resource {
    debugName;
    isResource = true;
    extent;
    graph;
    subsequents = new Set();
    suppliedBy = null;
    skipChecks = false;
    didUpdateSubscribers;
    constructor(extent, name) {
        this.extent = extent;
        this.graph = extent.graph;
        extent.addResource(this);
        if (name !== undefined) {
            this.debugName = name;
        }
        else {
            this.debugName = null;
        }
    }
    get order() {
        return { resource: this, type: LinkType.order };
    }
    get resource() {
        return this;
    }
    get type() {
        return LinkType.reactive;
    }
    toString() {
        let name = "Resource";
        if (this.debugName != null) {
            name = this.debugName + "(r)";
        }
        return name;
    }
    assertValidUpdater() {
        let graph = this.graph;
        let currentBehavior = graph.currentBehavior;
        let currentEvent = graph.currentEvent;
        if (currentBehavior == null && currentEvent == null) {
            let err = new Error("Resource must be updated inside a behavior or action.");
            err.resource = this;
            throw err;
        }
        if (this.skipChecks) {
            return;
        }
        if (this.suppliedBy && currentBehavior != this.suppliedBy) {
            let err = new Error("Supplied resource can only be updated by its supplying behavior.");
            err.resource = this;
            err.currentBehavior = currentBehavior;
            throw err;
        }
        if (this.suppliedBy == null && currentBehavior != null) {
            let err = new Error("Unsupplied resource can only be updated in an action.");
            err.resource = this;
            err.currentBehavior = currentBehavior;
            throw err;
        }
    }
    assertValidAccessor() {
        let graph = this.graph;
        let currentBehavior = graph.currentBehavior;
        if (currentBehavior != null && currentBehavior != this.suppliedBy && !currentBehavior.demands?.has(this)) {
            let err = new Error("Cannot access the value or event of a resource inside a behavior unless it is supplied or demanded.");
            err.resource = this;
            err.currentBehavior = currentBehavior;
            throw err;
        }
    }
    get justUpdated() {
        this.assertValidAccessor();
        return false;
    }
    subscribeToJustUpdated(callback) {
        return this._subscribeToJustUpdated({ extent: null, callback: callback });
    }
    _subscribeToJustUpdated(subscription) {
        // returns an unsubscribe callback that caller can call when no longer needed
        if (this.didUpdateSubscribers === undefined) {
            this.didUpdateSubscribers = new Set();
        }
        this.didUpdateSubscribers.add(subscription);
        return (() => {
            this.didUpdateSubscribers.delete(subscription);
        });
    }
    notifyJustUpdatedSubscribers() {
        if (this.didUpdateSubscribers !== undefined && this.didUpdateSubscribers.size > 0) {
            this.graph._notifyJustUpdatedSubscribers(this.didUpdateSubscribers);
        }
    }
}
export class Moment extends Resource {
    _happened = false;
    _happenedValue = undefined;
    _happenedWhen = null;
    get justUpdated() {
        this.assertValidAccessor();
        return this._happened;
    }
    get value() {
        this.assertValidAccessor();
        return this._happenedValue;
    }
    get event() {
        this.assertValidAccessor();
        return this._happenedWhen;
    }
    toString() {
        let name = "Moment";
        if (this.debugName != null) {
            name = (this.debugName + "(m)");
        }
        if (this._happenedValue !== undefined) {
            name = name + "=" + this._happenedValue;
        }
        if (this._happenedWhen !== null) {
            name = name + " : " + this._happenedWhen.sequence;
        }
        return name;
    }
    justUpdatedTo(value) {
        return this.justUpdated && this._happenedValue == value;
    }
    updateWithAction(value = undefined, debugName) {
        this.graph.action(() => {
            this.update(value);
        }, debugName);
        return;
    }
    update(value = undefined) {
        this.assertValidUpdater();
        this._happened = true;
        this._happenedValue = value;
        this._happenedWhen = this.graph.currentEvent;
        this.notifyJustUpdatedSubscribers();
        this.graph.resourceTouched(this);
        this.graph.trackTransient(this);
    }
    clear() {
        this._happened = false;
        this._happenedValue = undefined;
    }
}
export class State extends Resource {
    currentState;
    previousState = null;
    constructor(extent, initialState, name) {
        super(extent, name);
        this.currentState = { value: initialState, event: GraphEvent.initialEvent };
    }
    toString() {
        let name = "State";
        if (this.debugName != null) {
            name = (this.debugName + "(s)");
        }
        name = name + "=" + this.currentState.value;
        name = name + " : " + this.currentState.event.sequence;
        return name;
    }
    updateWithAction(newValue, debugName) {
        this.graph.action(() => {
            this.update(newValue);
        }, debugName);
        return;
    }
    update(newValue) {
        if (this.currentState.value === newValue) {
            return;
        }
        this.updateForce(newValue);
    }
    updateForce(newValue) {
        this.assertValidUpdater();
        this._updateForce(newValue);
    }
    _updateForce(newValue) {
        if (this.graph.currentEvent != null && this.currentState.event.sequence < this.graph.currentEvent?.sequence) {
            // captures trace as the value before any updates
            this.previousState = this.currentState;
        }
        this.currentState = { value: newValue, event: this.graph.currentEvent };
        this.notifyJustUpdatedSubscribers();
        this.graph.resourceTouched(this);
        this.graph.trackTransient(this);
    }
    clear() {
        this.previousState = null;
    }
    get value() {
        this.assertValidAccessor();
        return this.currentState.value;
    }
    get event() {
        this.assertValidAccessor();
        return this.currentState.event;
    }
    get trace() {
        if (this.currentState.event === this.graph.currentEvent) {
            return this.previousState;
        }
        else {
            return this.currentState;
        }
    }
    get traceValue() {
        return this.trace.value;
    }
    get traceEvent() {
        return this.trace.event;
    }
    get justUpdated() {
        this.assertValidAccessor();
        return this.currentState.event === this.graph.currentEvent;
    }
    justUpdatedTo(toState) {
        return this.justUpdated && this.currentState.value == toState;
    }
    justUpdatedFrom(fromState) {
        return this.justUpdated && this.traceValue == fromState;
    }
    justUpdatedToFrom(toState, fromState) {
        return this.justUpdatedTo(toState) && this.justUpdatedFrom(fromState);
    }
}
//# sourceMappingURL=resource.js.map